/** ZCT/eval.c--Created 071505 **/
#include "zct.h"
#include "functions.h"
#include "globals.h"
#include "eval.h"

BITBOARD (*fill_forward[2])(BITBOARD g, BITBOARD p) = { fill_up, fill_down };
BITBOARD (*smear_forward[2])(BITBOARD g) = { smear_up, smear_down };
/* The least and the greatest files that relevant pawns are on for pawn shelters. */
SQ_FILE shelter_file[2][2] = { { FILE_F, FILE_H }, { FILE_A, FILE_D } };

VALUE material_balance(void)
{
	COLOR color;
	PIECE piece;
	VALUE eval[2];

	for (color = WHITE; color <= BLACK; color++)
	{
		eval[color] = 0;
		for (piece = PAWN; piece < KING; piece++)
			eval[color] += piece_value[piece] * pop_count(board.piece_bb[piece] & board.color_bb[color]);
		DEBUG_EVAL(print("material[%C] = %V\n", color, eval[color]));
	}
	return eval[board.side_tm] - eval[board.side_ntm];
}

/**
evaluate():
Evaluates the board position.
Created 071505; last modified 020808
**/
VALUE evaluate(EVAL_BLOCK *eval_block)
{
	int r;
	BITBOARD pieces;
	COLOR color;
	EVAL_HASH_ENTRY *eval_hash_entry;
	PHASE phase;
	SQUARE square;
	VALUE eval_opening[2];
	VALUE eval_endgame[2];
	VALUE eval_temp;
	VALUE eval_temp_2;

//return material_balance();
	/* Look up this position in the eval hash table. */
	eval_hash_entry = &eval_hash_table[board.hashkey % eval_hash_size];
	eval_hash_probes++;
	if (eval_hash_entry && eval_hash_entry->hashkey == board.hashkey)
	{
		eval_hash_hits++;
		DEBUG_EVAL(print("eval hash hit = %V\n", eval_hash_entry->eval.eval));
		*eval_block = eval_hash_entry->eval;
		return eval_hash_entry->eval.eval;
	}

	eval_opening[WHITE] = eval_opening[BLACK] = 0;
	eval_endgame[WHITE] = eval_endgame[BLACK] = 0;
	/* Determine the game phase. */
	r = 0;
	r += 2 * MIN(board.piece_count[WHITE] + board.piece_count[BLACK], 16);				/* 32 */
	r += 8 * (CAN_CASTLE(board.castle_rights, WHITE) != 0);								/*  8 */
	r += 8 * (CAN_CASTLE(board.castle_rights, BLACK) != 0);								/*  8 */
	r += pop_count(board.piece_bb[PAWN]);												/* 16 */
	r += 3 * MIN(pop_count(board.piece_bb[KNIGHT] | board.piece_bb[BISHOP]), 6);		/* 18 */
	r += 5 * MIN(pop_count(board.piece_bb[ROOK]), 4);									/* 20 */
	r += 9 * MIN(pop_count(board.piece_bb[QUEEN]), 2);									/* 18 */
	r += pop_count(board.occupied_bb & (MASK_RANK(RANK_1) | MASK_RANK(RANK_8))) / 2;	/*  8 */

	phase = MAX_PHASE - r;
	eval_block->phase = phase;
	ASSERT(phase >= 0 && phase <= MAX_PHASE);
	DEBUG_EVAL(print("phase = %i/%i\n", phase, MAX_PHASE));

	/* First pass: Collect information about the board, evaluate simple terms. */
	for (color = WHITE; color <= BLACK; color++)
	{
		/* Material value and piece square tables. */
		pieces = board.color_bb[color] & ~board.piece_bb[KING];
		eval_temp = eval_temp_2 = 0;
		while (pieces)
		{
			square = first_square(pieces);
			CLEAR_BIT(pieces, square);
			eval_temp += piece_value[board.piece[square]] + piece_square_value[color][board.piece[square]][square];
			eval_temp_2 += piece_endgame_value[board.piece[square]];
		}
		DEBUG_EVAL(print("piece square[%C](op,eg) = %V %V\n", color, eval_temp, eval_temp_2));
		eval_block->material[color] = eval_temp;
		eval_opening[color] += eval_temp;
		eval_endgame[color] += eval_temp_2;

		/* Development. */
		eval_temp = 0;
		eval_temp += development_value[pop_count(board.color_bb[color] &
			(board.piece_bb[KNIGHT] | board.piece_bb[BISHOP]) & development_mask[color])];
		if (eval_temp < development_value[1] && board.piece[SQ_FLIP_COLOR(D1, color)] != QUEEN)
			eval_temp += early_queen_value;
		DEBUG_EVAL(print("development[%C] = %V\n", color, eval_temp));
		eval_opening[color] += eval_temp;
	}

	/* Evaluate pawn structure. */
	eval_temp = evaluate_pawns();
	DEBUG_EVAL(print("pawns = %V\n", eval_temp));
	eval_opening[board.side_tm] += eval_temp;
	eval_endgame[board.side_tm] += eval_temp;

	/* Second pass: evaluate more complex terms which need previously calculated information */
	for (color = WHITE; color <= BLACK; color++)
	{
		/* Evaluate knights. */
		pieces = board.color_bb[color] & board.piece_bb[KNIGHT];
		eval_temp = 0;
		while (pieces)
		{
			square = first_square(pieces);
			CLEAR_BIT(pieces, square);
			eval_temp += evaluate_knight(square, color);
		}
		DEBUG_EVAL(print("knights[%C] = %V\n", color, eval_temp));
		eval_opening[color] += eval_temp;
		eval_endgame[color] += eval_temp;
		/* Evaluate bishops. */
		pieces = board.color_bb[color] & board.piece_bb[BISHOP];
		eval_temp = 0;
		if (pop_count(pieces) >= 2)
			eval_temp += bishop_pair_value;
		while (pieces)
		{
			square = first_square(pieces);
			CLEAR_BIT(pieces, square);
			eval_temp += evaluate_bishop(square, color);
		}
		DEBUG_EVAL(print("bishops[%C] = %V\n", color, eval_temp));
		eval_opening[color] += eval_temp;
		eval_endgame[color] += eval_temp;
		/* Evaluate rooks. */
		pieces = board.color_bb[color] & board.piece_bb[ROOK];
		eval_temp = 0;
		while (pieces)
		{
			square = first_square(pieces);
			CLEAR_BIT(pieces, square);
			eval_temp += evaluate_rook(square, color);
		}
		DEBUG_EVAL(print("rooks[%C] = %V\n", color, eval_temp));
		eval_opening[color] += eval_temp;
		eval_endgame[color] += eval_temp;
		/* Evaluate queens. */
		pieces = board.color_bb[color] & board.piece_bb[QUEEN];
		eval_temp = 0;
		while (pieces)
		{
			square = first_square(pieces);
			CLEAR_BIT(pieces, square);
			eval_temp += evaluate_queen(square, color);
		}
		DEBUG_EVAL(print("queens[%C] = %V\n", color, eval_temp));
		eval_opening[color] += eval_temp;
		eval_endgame[color] += eval_temp;
		/* Evaluate king safety. */
		eval_temp = evaluate_king_safety(color);
		eval_block->king_safety[color] = eval_temp;
		DEBUG_EVAL(print("king safety[%C] = %V\n", color, eval_temp));
		eval_opening[color] += eval_temp;
		eval_endgame[color] += eval_temp / 2;

		/* Evaluate passed pawns. */
		eval_temp = evaluate_passed_pawns(color);
		DEBUG_EVAL(print("passed pawns[%C] = %V\n", color, eval_temp));
		eval_opening[board.side_tm] += eval_temp / 4;
		eval_endgame[board.side_tm] += eval_temp;
	}
	/* endgame eval */
	eval_temp = evaluate_endgame();
	DEBUG_EVAL(print("endgame eval = %V\n", eval_temp));
	eval_endgame[board.side_tm] += eval_temp;
	/* Side to move bonus. */
	eval_opening[board.side_tm] += side_tm_value;
	eval_endgame[board.side_tm] += side_tm_value;
	eval_temp = ((eval_opening[board.side_tm] - eval_opening[board.side_ntm]) * (MAX_PHASE - phase) +
		(eval_endgame[board.side_tm] - eval_endgame[board.side_ntm]) * phase) / MAX_PHASE;
	/* Store the evaluation in the hash table. */
	eval_hash_entry->hashkey = board.hashkey;
	eval_hash_entry->eval.eval = eval_temp;

	return eval_temp;
}

/**
evaluate_passed_pawns():
Evaluate any passed pawns that might be on the board.
Created 110506; last modified 020808
**/
VALUE evaluate_passed_pawns(COLOR color)
{
	int distance;
	int multiplier;
	BITBOARD pawns;
	BITBOARD friendly;
	BITBOARD rook_defend;
	BITBOARD attackers[6];
	SQUARE square;
	SQUARE front_square;
	VALUE eval_temp;

	pawns = board.pawn_hash_entry.passed_pawns & board.color_bb[color];
	eval_temp = 0;
	rook_defend = SHIFT_FORWARD(fill_forward[color]((board.piece_bb[ROOK] | board.piece_bb[QUEEN]) &
	board.color_bb[color], ~board.occupied_bb), color);
	while (pawns)
	{
		square = first_square(pawns);
		CLEAR_BIT(pawns, square);

		distance = 7 - RANK_OF(SQ_FLIP_COLOR(square, color));
		if (rook_defend & MASK(square))
			multiplier = 2;
		else
			multiplier = 1;
		front_square = square;
		attackers[KNIGHT] =	fill_attacks_knight(board.piece_bb[KNIGHT] & board.color_bb[COLOR_FLIP(color)]);
		attackers[BISHOP] = fill_attacks_bishop(board.piece_bb[BISHOP] & board.color_bb[COLOR_FLIP(color)], board.occupied_bb);
		while (TRUE)
		{
			front_square += pawn_dir[color];
			if (board.color[front_square] != EMPTY || (is_attacked(front_square, COLOR_FLIP(color)) &&
				!is_attacked(front_square, color)))
				break;
			else if (RANK_OF(SQ_FLIP_COLOR(front_square, color)) == RANK_8)
			{
				multiplier = 4;
				break;
			}
			multiplier = MIN(multiplier + 1, 3);
		}
		/* connected passed pawns */
		friendly = board.piece_bb[PAWN] & board.color_bb[color];
		if ((friendly | SHIFT_FORWARD(friendly, color)) & (SHIFT_LF(MASK(square)) | SHIFT_RT(MASK(square))))
		{
			eval_temp += connected_pp_value;
			multiplier++;
		}
		eval_temp += multiplier * passed_pawn_value[distance];
	}
	return eval_temp;
}

/**
evaluate_pawns():
Evaluates the pawn structure.
Created 090106; last modified 082607
**/
VALUE evaluate_pawns(void)
{
	int pawn_color_count;
	COLOR color;
	COLOR sq_color;
	BITBOARD temp;
	BITBOARD pawns[2];
	BITBOARD pawn_attacks_left[2];
	BITBOARD pawn_attacks_right[2];
	BITBOARD pawn_attacks[2];
	BITBOARD smear[2];
	BITBOARD controlled[2];
	BITBOARD safe_path[2];
	BITBOARD weak[2];
	PAWN_HASH_ENTRY *pawn_hash_entry;
	SHELTER shelter;
	SQ_RANK rank;
	SQ_FILE file;
	VALUE eval_temp;

	pawn_hash_entry = &pawn_hash_table[board.pawn_hash_entry.hashkey % pawn_hash_size];
	pawn_hash_probes++;
	if (pawn_hash_entry->hashkey && pawn_hash_entry->hashkey == board.pawn_hash_entry.hashkey)
	{
		pawn_hash_hits++;
		board.pawn_hash_entry = *pawn_hash_entry;
		return pawn_hash_entry->eval[board.side_tm] - pawn_hash_entry->eval[board.side_ntm];
	}

	board.pawn_hash_entry.eval[WHITE] = board.pawn_hash_entry.eval[BLACK] = 0;
	board.pawn_hash_entry.doubled_pawns = (BITBOARD)0;
	board.pawn_hash_entry.passed_pawns = (BITBOARD)0;
//	board.pawn_hash_entry.blocked_pawns = (BITBOARD)0;
	/* First pass */
	for (color = WHITE; color <= BLACK; color++)
	{
		pawns[color] = board.piece_bb[PAWN] & board.color_bb[color];
		pawn_attacks_left[color] = SHIFT_LF(SHIFT_FORWARD(pawns[color], color));
		pawn_attacks_right[color] = SHIFT_RT(SHIFT_FORWARD(pawns[color], color));
		pawn_attacks[color] = pawn_attacks_left[color] | pawn_attacks_right[color];
		smear[color] = smear_forward[color](pawns[color]);
		/* Open files */
		if (color == WHITE)
			board.pawn_hash_entry.open_files[WHITE] = (unsigned char)smear_down(smear[WHITE]);
		else
			board.pawn_hash_entry.open_files[BLACK] = (unsigned char)smear[BLACK];
		/* Doubled pawns */
		board.pawn_hash_entry.doubled_pawns |= pawns[color] & SHIFT_FORWARD(smear[color], color);
		eval_temp = pop_count(board.pawn_hash_entry.doubled_pawns & board.color_bb[color]) * doubled_pawn_value;
		DEBUG_EVAL(print("doubled pawns[%C] = %V\n", color, eval_temp));
		board.pawn_hash_entry.eval[color] += eval_temp;
		/* Count the number of pawns on each square color. */
		for (sq_color = WHITE; sq_color <= BLACK; sq_color++)
		{
			pawn_color_count = pop_count(board.piece_bb[PAWN] & board.color_bb[color] & SQ_COLOR_MASK(sq_color));
			eval_temp = pawn_color_count - pop_count(board.piece_bb[PAWN] & board.color_bb[color]);
			board.pawn_hash_entry.bishop_color_value[color][sq_color] = eval_temp * eval_temp;
		}
	}
	/* Second pass */
	for (color = WHITE; color <= BLACK; color++)
	{
		/* Passed pawns */
		if (color == WHITE)
			board.pawn_hash_entry.passed_pawns |= ~(SHIFT_DL(smear[BLACK]) |
			SHIFT_DR(smear[BLACK]) | smear[BLACK]) & pawns[WHITE];
		else
			board.pawn_hash_entry.passed_pawns |= ~(SHIFT_UL(smear[WHITE]) |
			SHIFT_UR(smear[WHITE]) | smear[WHITE]) & pawns[BLACK];
		DEBUG_EVAL(print("passed pawns[%C]:\n%I", color, board.pawn_hash_entry.passed_pawns[color]));
		/* Find squares with more pawns of this color attacking it than the opposite color. */
		controlled[color] = (pawn_attacks[color] & ~pawn_attacks[COLOR_FLIP(color)]) |
			((pawn_attacks_left[color] & pawn_attacks_right[color]) &
			~(pawn_attacks_left[COLOR_FLIP(color)] & pawn_attacks_right[COLOR_FLIP(color)]));
		/* Find pawns blocked by pawns of the opposite color. */
//		board.pawn_hash_entry.blocked_pawns |= SHIFT_FORWARD(pawns[COLOR_FLIP(color)], COLOR_FLIP(color)) & pawns[color];
		/* Remember squares not controlled by opponent pawns. */
		board.pawn_hash_entry.not_attacked[color] = ~pawn_attacks[COLOR_FLIP(color)];
		/* Evaluate king-pawn shelters on king and queen side. */
		for (shelter = KING_SIDE; shelter <= QUEEN_SIDE; shelter++)
		{
			eval_temp = 1;
			for (file = shelter_file[shelter][0]; file <= shelter_file[shelter][1]; file++)
			{
				if (!(pawns[color] & MASK_FILE(file)))
					eval_temp *= 6;
				else
				{
					if (color == WHITE)
						rank = RANK_OF(first_square(pawns[color] & MASK_FILE(file))) - RANK_2;
					else
						rank = RANK_7 - RANK_OF(last_square(pawns[color] & MASK_FILE(file)));
					eval_temp *= MIN(1 << rank, 6);
				}
			}
			eval_temp = MIN(eval_temp, 41) - 1;
			board.pawn_hash_entry.king_shelter_value[color][shelter] = king_shelter_value[QUEEN_SIDE][eval_temp];
		}
	}
	/* Third pass */
	for (color = WHITE; color <= BLACK; color++)
	{
		/* Find paths that pawns can safely advance through. */
		safe_path[color] = fill_forward[color](pawns[color], ~(board.piece_bb[PAWN] | controlled[COLOR_FLIP(color)]));
		/* Find weak pawns. */
		temp = SHIFT_FORWARD(fill_forward[color](pawns[color], safe_path[color]), color);
		weak[color] = SHIFT_FORWARD(~board.piece_bb[PAWN] & controlled[COLOR_FLIP(color)], COLOR_FLIP(color)) &
			~(SHIFT_LF(temp) | SHIFT_RT(temp)) & pawns[color];

		eval_temp = pop_count(weak[color]) * weak_pawn_value;
		DEBUG_EVAL(print("weak pawns[%C] = %V\n", color, eval_temp));
		board.pawn_hash_entry.eval[color] += eval_temp;
	}

	*pawn_hash_entry = board.pawn_hash_entry;
	return board.pawn_hash_entry.eval[board.side_tm] - board.pawn_hash_entry.eval[board.side_ntm];
}

/**
evaluate_knight():
Evaluates a knight on the given square.
Created 091306; last modified 091406
**/
VALUE evaluate_knight(SQUARE square, COLOR color)
{
	VALUE r = 0;

	r += evaluate_mobility(knight_moves_bb[square], KNIGHT, color, square);
	return r;
}

/**
evaluate_bishop():
Evaluates a bishop on the given square.
Created 091306; last modified 111207
**/
VALUE evaluate_bishop(SQUARE square, COLOR color)
{
	BITBOARD attacks;
	VALUE r;

	r = 0;
	/* pawn color value */
	r += board.pawn_hash_entry.bishop_color_value[color][SQ_COLOR(square)];
	/* trapped on A7/H7 with pawn blocking */
	if (square == SQ_FLIP_COLOR(A7, color) && board.piece[SQ_FLIP_COLOR(B6, color)] == PAWN &&
		board.color[SQ_FLIP_COLOR(B6, color)] == COLOR_FLIP(color))
		r += bishop_trapped_value;
	if (square == SQ_FLIP_COLOR(H7, color) && board.piece[SQ_FLIP_COLOR(G6, color)] == PAWN &&
		board.color[SQ_FLIP_COLOR(G6, color)] == COLOR_FLIP(color))
		r += bishop_trapped_value;
	/* mobility */
	attacks = BISHOP_ATTACKS(square, board.occupied_bb);
	r += evaluate_mobility(attacks, BISHOP, color, square);

	return r;
}

/**
evaluate_rook():
Evaluates a rook on the given square.
Created 110106; last modified 111207
**/
VALUE evaluate_rook(SQUARE square, COLOR color)
{
	BITBOARD attacks;
	SQ_FILE f;
	VALUE r = 0;

	/* rook on seventh */
	r += rook_on_seventh_value[RANK_OF(SQ_FLIP_COLOR(square, color)) == RANK_7];
	/* rook on open file */
	f = FILE_OF(square);
	/* For the open file value, we need two bits, for whether the file is open for each side. */
	r += rook_open_file_value[((board.pawn_hash_entry.open_files[color] >> f) & 1) |
		((board.pawn_hash_entry.open_files[COLOR_FLIP(color)] >> (f - 1)) & 2)];
	/* Mobility: we ignore friendly rooks and queens because they can be "attacked through". */
	attacks = ROOK_ATTACKS(square, board.occupied_bb & ~((board.piece_bb[ROOK] | board.piece_bb[QUEEN]) & board.color_bb[color]));
	r += evaluate_mobility(attacks, ROOK, color, square);

	return r;
}

/**
evaluate_queen():
Evaluates a queen on the given square.
Created 110106; last modified 020808
**/
VALUE evaluate_queen(SQUARE square, COLOR color)
{
	BITBOARD attacks;
	SQ_FILE f;
	VALUE r = 0;

	/* queen on seventh */
	r += rook_on_seventh_value[RANK_OF(SQ_FLIP_COLOR(square, color)) == RANK_7];
	/* queen on open file */
	f = FILE_OF(square);
	/* For the open file value, we need two bits, for whether the file is open for each side. */
	r += rook_open_file_value[((board.pawn_hash_entry.open_files[color] >> f) & 1) |
		((board.pawn_hash_entry.open_files[COLOR_FLIP(color)] >> (f - 1)) & 2)];
	/* Mobility: we ignore friendly rooks and bishops (in the right direction) and queens because they can be "attacked through"  */
	attacks = ROOK_ATTACKS(square, board.occupied_bb & ~((board.piece_bb[ROOK] | board.piece_bb[QUEEN]) & board.color_bb[color])) |
		BISHOP_ATTACKS(square, board.occupied_bb & ~((board.piece_bb[BISHOP] | board.piece_bb[QUEEN]) & board.color_bb[color]));
	r += evaluate_mobility(attacks, QUEEN, color, square) / 2;

	return r;
}

/**
evaluate_mobility():
Takes a given attack set and computes an evaluation based on what the piece attacks.
Created 102706; last modifed 092907
**/
VALUE evaluate_mobility(BITBOARD attacks, PIECE piece, COLOR color, SQUARE square)
{
	VALUE r = 0;

	r = mobility_value[piece][pop_count(attacks & ~board.color_bb[color])] +
		safe_mobility_value[piece][pop_count(attacks & board.pawn_hash_entry.not_attacked[color])];
	return r;
}

/**
evaluate_endgame():
Evaluates all matters pertaining to the endgame.
Created 100807; last modified 121807
**/
VALUE evaluate_endgame(void)
{
	BITBOARD bb_temp;
	COLOR color;
	SQUARE square;
	VALUE eval_temp[2];

	for (color = WHITE; color <= BLACK; color++)
	{
		eval_temp[color] = 0;
		/* Evaluate single bishops, trying to force the opponent king into the appropriate corner. */
		if (pop_count(board.piece_bb[BISHOP] & board.color_bb[color]) == 1)
		{
			square = first_square(board.piece_bb[BISHOP] & board.color_bb[color]);
			/* Flip the opponent's king square relative to the bishop square color. */
			eval_temp[color] += king_bishop_square_value[SQ_FLIP_COLOR(board.king_square[COLOR_FLIP(color)], SQ_COLOR(square))];
		}
		/* Evaluate how rooks restrict king movement. */
		if (board.piece_bb[ROOK] & board.color_bb[color])
		{
			bb_temp = fill_attacks_rook(board.piece_bb[ROOK] & board.color_bb[color], board.occupied_bb) |
				(board.piece_bb[ROOK] & board.color_bb[color]);
			eval_temp[color] += pop_count(flood_fill_king(MASK(board.king_square[COLOR_FLIP(color)]), bb_temp, 0));
		}
		/* Evaluate king centralization. */
		eval_temp[color] += king_endgame_square_value[board.king_square[color]];
	}
	return eval_temp[board.side_tm] - eval_temp[board.side_ntm];
}

/**
can_mate():
This function returns false if, based on the material configuration, it is impossible for
the given side to force mate.
Created 011508; last modified 020208
**/
BOOL can_mate(COLOR color)
{
	if (board.piece_bb[PAWN] & board.color_bb[color])
		return TRUE;
	if (board.piece_bb[ROOK] & board.color_bb[color])
		return TRUE;
	if (board.piece_bb[QUEEN] & board.color_bb[color])
		return TRUE;
	if (pop_count(board.piece_bb[KNIGHT] & board.color_bb[color]) > 1)
		return TRUE;
	if (pop_count(board.piece_bb[BISHOP] & board.color_bb[color]) > 1)
		return TRUE;
	if (pop_count((board.piece_bb[KNIGHT] | board.piece_bb[BISHOP]) & board.color_bb[color]) > 1)
		return TRUE;
	return FALSE;
}

/**
evaluate_king_safety():
Evaluates the overall safety of the given color's king.
Created 110606; last modified 020808
**/
VALUE evaluate_king_safety(COLOR color)
{
	int r;
	BITBOARD king_area;
	BITBOARD bishop_attacks;
	BITBOARD rook_attacks;
	BITBOARD pawn_attacks;
	BITBOARD empty;
	BOOL king_is_safe;
	SHELTER shelter;
	SQUARE square;
	SQ_FILE file;
	VALUE eval_temp;
	VALUE eval;
	VALUE attack_value;

	eval = 0;
	square = board.king_square[color];
	empty = ~board.occupied_bb;
	king_area = king_moves_bb[square] | mask_bb[square];
	/* piece attacks */
	r = 0;
	pawn_attacks = SHIFT_FORWARD(board.piece_bb[PAWN] & board.color_bb[COLOR_FLIP(color)], COLOR_FLIP(color));
	pawn_attacks = SHIFT_LF(pawn_attacks) | SHIFT_RT(pawn_attacks);
    bishop_attacks = SHIFT_UL(fill_up_left(king_area, empty)) | SHIFT_UR(fill_up_right(king_area, empty))
        | SHIFT_DR(fill_down_right(king_area, empty)) | SHIFT_DL(fill_down_left(king_area, empty));
    rook_attacks = SHIFT_UP(fill_up(king_area, empty)) | SHIFT_DN(fill_down(king_area, empty))
        | SHIFT_LF(fill_left(king_area, empty)) | SHIFT_RT(fill_right(king_area, empty));
	r += king_safety_att_weight[PAWN] * pop_count(pawn_attacks & king_area);
	r += king_safety_att_weight[BISHOP] * pop_count(bishop_attacks & board.piece_bb[BISHOP] & board.color_bb[COLOR_FLIP(color)]);
	r += king_safety_att_weight[ROOK] * pop_count(rook_attacks & board.piece_bb[ROOK] & board.color_bb[COLOR_FLIP(color)]);
	r += king_safety_att_weight[QUEEN] * pop_count((bishop_attacks | rook_attacks) & board.piece_bb[QUEEN] &
		board.color_bb[COLOR_FLIP(color)]);
	r = MIN(MAX(r, 0), 40);
	eval_temp = attack_value = king_safety_att_value[r];
	eval += eval_temp;
	DEBUG_EVAL(print("king safety: piece attacks[%C]=%V\n", color, eval_temp));
	/* blocking */
#if 0
	r = 0;
	r += king_safety_block_weight[0] * pop_count(king_area & board.piece_bb[PAWN] & board.color_bb[color]);
	r += king_safety_block_weight[1] * pop_count(king_area & (board.piece_bb[KNIGHT] | board.piece_bb[BISHOP] |
		board.piece_bb[ROOK] | board.piece_bb[QUEEN]) & board.color_bb[color]);
	r = MIN(r, 20);
	eval_temp = king_safety_block_value[r];
	eval += eval_temp;
	DEBUG_EVAL(print("king safety: blocking[%C]=%V\n", color, eval_temp));
	/* defending */
	r = 0;
	king_area = king_moves_bb[square] & board.color_bb[color];
	pawn_attacks = SHIFT_FORWARD(board.piece_bb[PAWN] & board.color_bb[color], color);
	pawn_attacks = SHIFT_LF(pawn_attacks) | SHIFT_RT(pawn_attacks);
    bishop_attacks = SHIFT_UL(fill_up_left(king_area, empty)) | SHIFT_UR(fill_up_right(king_area, empty))
        | SHIFT_DR(fill_down_right(king_area, empty)) | SHIFT_DL(fill_down_left(king_area, empty));
    rook_attacks = SHIFT_UP(fill_up(king_area, empty)) | SHIFT_DN(fill_down(king_area, empty))
        | SHIFT_LF(fill_left(king_area, empty)) | SHIFT_RT(fill_right(king_area, empty));
	r += king_safety_def_weight[PAWN] * pop_count(pawn_attacks & king_area);
	r += king_safety_def_weight[BISHOP] * pop_count(bishop_attacks & board.piece_bb[BISHOP] & board.color_bb[color]);
	r += king_safety_def_weight[ROOK] * pop_count(rook_attacks & board.piece_bb[ROOK] & board.color_bb[color]);
	r += king_safety_def_weight[QUEEN] * pop_count((bishop_attacks | rook_attacks) & board.piece_bb[QUEEN] & board.color_bb[color]);
	r = MIN(r, 40);
	eval_temp = (attack_value - king_safety_att_value[40]) * king_safety_def_percentage[r] / 100;
	eval += eval_temp;
	DEBUG_EVAL(print("king safety: defense factor[%C]=%i%%\n", color, king_safety_def_percentage[r]));
	DEBUG_EVAL(print("king safety: defense score[%C]=%V\n", color, eval_temp));
	DEBUG_EVAL(print("king safety: dynamic score[%C]=%V\n", color, eval));
	
#endif
	/* king shelter */
	eval_temp = 0;
	king_is_safe = FALSE;
	file = FILE_OF(square);
	for (shelter = KING_SIDE; shelter <= QUEEN_SIDE; shelter++)
	{
		if (file >= shelter_file[shelter][0] && file <= shelter_file[shelter][1])
		{
			eval_temp += board.pawn_hash_entry.king_shelter_value[color][shelter];
			king_is_safe = TRUE;
		}
	}
	DEBUG_EVAL(print("king safety: king_is_safe[%C]=%i\n", color, king_is_safe));
	if (!king_is_safe)
	{
		if (!CAN_CASTLE(board.castle_rights, color))
			eval_temp += lost_castling_value[1];
		else if (!CAN_CASTLE_KS(board.castle_rights, color) || !CAN_CASTLE_QS(board.castle_rights, color))
			eval_temp += lost_castling_value[0];
		/* Penalize wrecking the pawn structure before castling. */
		for (shelter = KING_SIDE; shelter <= QUEEN_SIDE; shelter++)
			eval_temp += board.pawn_hash_entry.king_shelter_value[color][shelter] / 2;
	}
	/* If the king is safe but got there without castling, we need to check for trapped rooks. */
	else if (RANK_OF(SQ_FLIP_COLOR(square, color)) == RANK_1)
	{
		if (trapped_rook_mask[color][file] & board.piece_bb[ROOK] & board.color_bb[color])
			eval_temp += trapped_rook_value;
	}
	eval += eval_temp;
	DEBUG_EVAL(print("king safety: shelter score[%C]=%V\n", color, eval_temp));
	return eval;
}
