/** ZCT/smp.c--Created 081305 **/
#include "zct.h"
#include "functions.h"
#include "globals.h"
#include "smp.h"

#ifdef SMP

SMP_BLOCK *smp_block = NULL;
SPLIT_POINT *split_point = NULL;
SMP_DATA *smp_data = NULL;
TREE_BLOCK *tree_block = NULL;
int sem_key = 0;

/**
initialize_smp():
Initializes the smp functionality for the given number of processes.
Created 081305; last modified 110907
**/
void initialize_smp(int procs)
{
	int x;
	int y;
	int smp_block_size;
	int split_point_size;
	int tree_block_size;
	int size;
	void *mem;
#ifdef ZCT_USE_SEMAPHORES
	int sem_count;
	union semun sem_init;
#endif

	if (smp_data != NULL)
		smp_cleanup();
	else if (atexit(smp_cleanup) == -1)
	{
		smp_cleanup();
		perror("atexit failed");
		exit(1);
	}
//	signal(SIGINT, smp_cleanup_sig);
	signal(SIGTERM, smp_cleanup_sig);
//	signal(SIGCHLD, smp_cleanup_sig);

	process_count = procs;

	/* Create the shared memory */
	smp_block_size = sizeof(SMP_BLOCK) * procs;
	split_point_size = sizeof(SPLIT_POINT) * MAX_SPLIT_POINTS;
	tree_block_size = sizeof(TREE_BLOCK) * procs * procs;
	size = smp_block_size + split_point_size + tree_block_size + sizeof(SMP_DATA);
	mem = shared_alloc(size);

	/* Allocate the shared memory to the various data structures needed. */
	smp_block = (SMP_BLOCK *)mem;
	split_point = (SPLIT_POINT *)(smp_block + smp_block_size);
	tree_block = (TREE_BLOCK *)(smp_block + smp_block_size + split_point_size);
	smp_data = (SMP_DATA *)(smp_block + smp_block_size + split_point_size + tree_block_size);
	/* Initialize the data. */
	/* SMP data */
	smp_data->split_count = 0;
	smp_data->split_id = 0;
	smp_data->return_flag = FALSE;
	/* smp blocks */
	for (x = 0; x < procs; x++)
	{
		smp_block[x].id = x;
		smp_block[x].idle = FALSE;
		smp_block[x].message_count = 0;
		smp_block[x].input = 0;
		smp_block[x].output = 0;
	}
	/* split points */
	for (x = 0; x < MAX_SPLIT_POINTS; x++)
	{
		split_point[x].active = FALSE;
		split_point[x].child_count = 0;
		for (y = 0; y < MAX_CPUS; y++)
		{
			split_point[x].update[y] = FALSE;
			split_point[x].is_child[y] = FALSE;
		}
	}
	/* tree blocks */
	for (x = 0; x < procs * procs; x++)
		tree_block[x].used = FALSE;
	/* main processor's smp info */
	board.id = 0;
	board.split_ply = board.split_ply_stack;
	board.split_ply_stack[0] = -1;
	board.split_point = board.split_point_stack;
	board.split_point_stack[0] = NULL;

#ifdef ZCT_USE_SEMAPHORES
	/* Initialize the semaphore set for POSIX machines without native spinlocks. */
	sem_count = procs * 2 + 2;
	if ((sem_key = semget(IPC_PRIVATE, sem_count, 0660 | IPC_CREAT)) == -1)
	{
		perror("semget failed");
		exit(1);
	}
	sem_init.val = 1;
	for (x = 0; x < sem_count; x++)
		semctl(sem_key, x, SETVAL, sem_init);
	/* Assign each lock a semaphore from the set */
	smp_data->io_lock = 0;
	smp_data->lock = 1;
	for (x = 0; x < procs; x++)
	{
		smp_block[x].lock = 2 * x + 2;
		smp_block[x].input_lock = 2 * x + 3;
	}
	for (x = 0; x < MAX_SPLIT_POINTS; x++)
		split_point[x].lock = x + 2 * procs + 2;
#else
	/* Initialize the OSX spin locks. These are much simpler, as they are just integers! */
	LOCK_INIT(smp_data->io_lock);
	LOCK_INIT(smp_data->lock);
	for (x = 0; x < procs; x++)
	{
		LOCK_INIT(smp_block[x].lock);
		LOCK_INIT(smp_block[x].input_lock);
	}
	for (x = 0; x < MAX_SPLIT_POINTS; x++)
		LOCK_INIT(split_point[x].lock);
#endif

	/* Now start the child processes. */
	for (x = 1; x < procs; x++)
	{
		/* Child process. */
		if ((y = fork()) == 0)
		{
			board.id = x;
			idle_loop(x);
		}
		/* Parent process. */
		else
		{
			smp_block[x].pid = y;
			smp_tell(x, SMP_INIT, 0);
			smp_tell(x, SMP_SEARCH, 0);
		}
	}
}

/**
shared_alloc():
Allocate memory to be shared among multiple processors.
Created 123107; last modified 123107
**/
void *shared_alloc(BITBOARD size)
{
	void *mem;

#ifdef ZCT_POSIX
	if ((mem = mmap(0, size, PROT_READ | PROT_WRITE, MAP_ANON | MAP_SHARED /*| MAP_HASSEMAPHORE*/, -1, 0)) == MAP_FAILED)
	{
		perror("mmap failed");
		exit(1);
	}
#elif defined(ZCT_WINDOWS)
	HANDLE map_file;

	map_file = CreateFileMapping(INVALID_HANDLE_VALUE, NULL, PAGE_READWRITE, 0, size, "ZCTMAPFILE");
	if (map_file == NULL)
	{
		perror("CreateFileMapping failed");
		exit(1);
	}
	mem = (void *)MapViewOfFile(map_file, FILE_MAP_ALL_ACCESS, 0, 0, size);
#endif
	return mem;
}

#ifdef ZCT_WINDOWS
/**
fork():
This extremely ugly function is the Windows equivalent of fork(). It was contributed by Dann Corbit.
Note that the ugliness is not his fault!
Created 022408; last modifed 022408
**/
int fork(void) 
{ 
	char szPath[FILENAME_MAX]; 
	char szCommandLine[1024]; 
	PROCESS_INFORMATION pi; 
	STARTUPINFO	startup_info; 

	startup_info.cb = sizeof(STARTUPINFO); 
	startup_info.lpReserved = NULL; 
	startup_info.lpDesktop = NULL; 
	startup_info.lpTitle = NULL; 
	startup_info.dwX = 0; 
	startup_info.dwY = 0; 
	startup_info.dwXSize = 0; 
	startup_info.dwYSize = 0; 
	startup_info.cbReserved2 = 0; 
	startup_info.lpReserved2 = NULL; 
	startup_info.dwFlags = STARTF_USESHOWWINDOW; 
	startup_info.wShowWindow = SW_HIDE; 

	szCommandLine[0] = '\0'; 

	GetModuleFileName(NULL, szPath, sizeof(szPath) - 1); 

	sprint(szCommandLine, sizeof(szCommandLine), "\"%s\"", szPath); 
	if (CreateProcess(NULL, szCommandLine, NULL, NULL, TRUE, 0, NULL, NULL, &startup_info, &pi) == 0) 
		return 0; 

	return pi.dwProcessId;
} 
#endif 

/**
idle_loop():
Loop the child processes while waiting for work.
Created 081405; last modified 082906
**/
void idle_loop(int id)
{
	while (TRUE)
	{
		while (smp_block[id].input == 0)
			;
		SMP_DEBUG(print("cpu %i got input %i\n", id, smp_block[id].input));
		switch (smp_block[id].input)
		{
			case SMP_INIT:
				initialize_board(NULL);
				smp_block[id].input = 0;
				smp_block[id].output = SMP_DONE;
				break;
			case SMP_SEARCH:
				initialize_board(NULL);
				board.search_stack[0].search_state = SEARCH_CHILD_RETURN;
				board.search_stack[1].search_state = SEARCH_PARK;
				smp_block[id].input = 0;
				smp_block[id].output = SMP_DONE;
				search(&board.search_stack[1]);
				print("CPU %i returned!!\n", id);
				exit(0);
				break;
			case SMP_PERFT:
				smp_block[id].input = 0;
				smp_block[id].output = SMP_DONE;
				break;
		}
	}
}

/**
smp_tell():
Sets the appropriate locks, sends a message to the given processor, and waits for the response.
Note that this function is synchronous, that is, the sending processor needs a response before
it can continue searching. The return value is the data sent back to the processor.
Created 082606; last modified 012208
**/
int smp_tell(int id, SMP_INPUT input, int data)
{
	int r;

	LOCK(smp_block[id].input_lock);
	smp_block[id].data = data;
	smp_block[id].output = 0;
	smp_block[id].input = input;
	while (smp_block[id].output == 0 && !smp_data->stop && !smp_data->return_flag)
	{
		/* Check if anyone is telling us to split too... */
		if (smp_block[board.id].input == SMP_SPLIT)
		{
			smp_block[board.id].input = 0;
			smp_block[board.id].data = -1;
			smp_block[board.id].output = SMP_DONE;
		}
	}
	/* Check for error: the search stopped or completed. */
	if (smp_block[id].output == 0)
	{
		smp_block[id].input = 0;
		UNLOCK(smp_block[id].input_lock);
		return -1;
	}
	smp_block[id].output = 0;
	smp_block[id].input = 0;
	r = smp_block[id].data;
	UNLOCK(smp_block[id].input_lock);
	return r;
}

/**
smp_message():
Sends a message to a given processor. Since these messages are asynchronous, and many processors
could be sending many messages at a time, there is a message queue. This goes in FIFO order.
Created 081007; last modified 012408
**/
void smp_message(int id, SMP_MESSAGE_TYPE message, int data)
{
	LOCK(smp_block[id].lock);
	/* As a worst case scenario, when the message queue fills up, we just return. */
	if (smp_block[id].message_count >= MAX_MESSAGES)
	{
		UNLOCK(smp_block[id].lock);
		return;
	}
	smp_block[id].message_queue[smp_block[id].message_count].message = message;
	smp_block[id].message_queue[smp_block[id].message_count].data = data;
	smp_block[id].message_count++;
	UNLOCK(smp_block[id].lock);
}

/**
smp_copy_to():
In a parallel search, copies the board state (moves etc.) to shared memory on a split.
Created 081706; last modified 022408
**/
void smp_copy_to(SPLIT_POINT *sp_to, GAME_ENTRY *ge, SEARCH_BLOCK *sb)
{
	int x;
	BOARD *to;
	EVAL_BLOCK *e_from;
	EVAL_BLOCK *e_to;
	GAME_ENTRY *game_p;
	MOVE *m_from;
	MOVE *m_to;
	SEARCH_BLOCK *s_from;
	SEARCH_BLOCK *s_to;

	to = &sp_to->board;

	/* Copy the move history. */
	game_p = root_entry;
	m_to = sp_to->move_path;
	while (game_p < ge)
		*m_to++ = game_p++->move;
	sp_to->last_move = m_to;

	/* Copy the eval stack. */
	e_from = board.eval_stack;
	e_to = to->eval_stack;
	while (e_from <= board.eval_stack + sb->ply)
		*e_to++ = *e_from++;
	/* Copy the move stack. */
	m_from = board.move_stack;
	m_to = to->move_stack;
	while (m_from < sb->last_move)
		*m_to++ = *m_from++;

	/* Copy threat moves. */
	for (x = 0; x < MAX_PLY; x++)
		to->threat_move[x] = board.threat_move[x];
	/* Copy the PVs. */
	for (x = 0; x <= sb->ply; x++)
		copy_pv(to->pv_stack[x], board.pv_stack[x]);

	/* Copy the search state. */
	s_from = board.search_stack + 1;
	s_to = to->search_stack + 1;
	while (s_from <= sb)
	{
		*s_to = *s_from;
		s_to->first_move = s_from->first_move - board.move_stack + to->move_stack;
		s_to->next_move = s_from->next_move - board.move_stack + to->move_stack;
		s_to->last_move = s_from->last_move - board.move_stack + to->move_stack;
		s_to++;
		s_from++;
	}
	
	/* Copy the split point information. */
	for (x = 0; x <= board.split_ply - board.split_ply_stack; x++)
		to->split_ply_stack[x] = board.split_ply_stack[x];
	to->split_ply = board.split_ply - board.split_ply_stack + to->split_ply_stack;

	for (x = 0; x <= board.split_point - board.split_point_stack; x++)
		to->split_point_stack[x] = board.split_point_stack[x];
	to->split_point = board.split_point - board.split_point_stack + to->split_point_stack;
}

/**
smp_copy_from():
In a parallel search, copies the board state (moves etc.) to shared memory on a split.
Created 022308; last modified 022408
**/
void smp_copy_from(SPLIT_POINT *sp_from, SEARCH_BLOCK *sb)
{
	int x;
	BOARD *from;
	EVAL_BLOCK *e_from;
	EVAL_BLOCK *e_to;
	MOVE *m_from;
	MOVE *m_to;
	SEARCH_BLOCK *s_from;
	SEARCH_BLOCK *s_to;

	from = &sp_from->board;

	/* Copy the move history, and make the moves. */
	m_from = sp_from->move_path;
	while (m_from < sp_from->last_move)
		make_move(*m_from++);

	/* Copy the eval stack. */
	e_from = from->eval_stack;
	e_to = board.eval_stack;
	while (e_from <= from->eval_stack + sb->ply)
		*e_to++ = *e_from++;
	/* Copy the move stack. */
	m_from = from->move_stack;
	m_to = board.move_stack;
	while (m_from < sb->last_move)
		*m_to++ = *m_from++;

	/* Copy threat moves. */
	for (x = 0; x < MAX_PLY; x++)
		board.threat_move[x] = from->threat_move[x];
	/* Copy the PVs. */
	for (x = 0; x <= sb->ply; x++)
		copy_pv(board.pv_stack[x], from->pv_stack[x]);

	/* Copy the search state. */
	s_from = from->search_stack + 1;
	s_to = board.search_stack + 1;
	while (s_from <= sb)
	{
		*s_to = *s_from;
		s_to->first_move = s_from->first_move - from->move_stack + board.move_stack;
		s_to->next_move = s_from->next_move - from->move_stack + board.move_stack;
		s_to->last_move = s_from->last_move - from->move_stack + board.move_stack;
		s_to++;
		s_from++;
	}
	
	/* Copy the split point information. */
	for (x = 0; x <= from->split_ply - from->split_ply_stack; x++)
		board.split_ply_stack[x] = from->split_ply_stack[x];
	board.split_ply = from->split_ply - from->split_ply_stack + board.split_ply_stack;

	for (x = 0; x <= from->split_point - from->split_point_stack; x++)
		board.split_point_stack[x] = from->split_point_stack[x];
	board.split_point = from->split_point - from->split_point_stack + board.split_point_stack;
}

/*
smp_copy_root():
At the beginning of a search, copy the root position so that we can coordinate the split points around it.
Created 022308; last modified 022308
**/
void smp_copy_root(BOARD *to, BOARD *from)
{
	int x;
	GAME_ENTRY *game_p;

	for (x = 0; x < 6; x++)
		to->piece_bb[x] = from->piece_bb[x];
	for (x = 0; x < 2; x++)
		to->color_bb[x] = from->color_bb[x];
	to->occupied_bb = from->occupied_bb;
	for (x = A1; x < OFF_BOARD; x++)
	{
		to->piece[x] = from->piece[x];
		to->color[x] = from->color[x];
	}
	for (x = 0; x < 2; x++)
	{
		to->king_square[x] = from->king_square[x];
		to->piece_count[x] = from->piece_count[x];
	}
	to->side_tm = from->side_tm;
	to->side_ntm = from->side_ntm;
	to->ep_square = from->ep_square;
	to->castle_rights = from->castle_rights;
	to->fifty_count = from->fifty_count;
	to->hashkey = from->hashkey;
	to->pawn_hash_entry.hashkey = from->pawn_hash_entry.hashkey;
	to->move_number = from->move_number;

	game_p = from->game_stack;
	to->game_entry = to->game_stack;
	while (game_p < from->game_entry)
		*to->game_entry++ = *game_p++;
}

/**
move_list_get():
In a split point, before a move is selected, the split point's move list is copied from
shared memory into the local move stack. Note that the lock is left locked, to be unlocked by put().
Created 082606; last modified 020608
**/
void move_list_get(SEARCH_BLOCK *sb)
{
	MOVE *from;
	MOVE *to;
	SPLIT_POINT *sp;

	sp = *board.split_point;
	LOCK(sp->lock);
	sb->select_state = sp->sb->select_state;
	to = sb->next_move;
	from = sp->move_list;
	while (*from)
		*to++ = *from++;
	sb->last_move = to;
}

/**
move_list_put():
In a split point, after a move is selected, the updated move list is copied back to shared memory.
Created 082606; last modified 020608
**/
void move_list_put(SEARCH_BLOCK *sb)
{
	MOVE *from;
	MOVE *to;
	SPLIT_POINT *sp;

	sp = *board.split_point;
	to = sp->move_list;
	from = sb->next_move;
	while (from < sb->last_move)
		*to++ = *from++;
	*to = NULL_MOVE;
	sp->sb->select_state = sb->select_state;
	UNLOCK(sp->lock);
}


/**
handle_messages():
During the search, if we receive a message from another processor, we drop into this function.
While we are here we process every message there is in the queue before releasing the lock.
Created 081207; last modified 020808
**/
void handle_messages(SEARCH_BLOCK **sb)
{
	int message;
	SMP_MESSAGE message_queue[MAX_MESSAGES];
	int message_count;

	LOCK(smp_block[board.id].lock);
//	SMP_DEBUG(print("cpu %i handling %i messages:", board.id, smp_block[board.id].message_count));
	/* Copy the message queue to local memory, so that we can process them without tying up the other
		processors waiting to message it (which could cause a deadlock). */
	message_count = smp_block[board.id].message_count;
	for (message = 0; message < smp_block[board.id].message_count; message++)
	{
		message_queue[message] = smp_block[board.id].message_queue[message];
//		SMP_DEBUG(print("(%i, %i) ", message_queue[message].message, message_queue[message].data));
	}
//	SMP_DEBUG(print("\n"));
	/* Reset the message queue. */
	smp_block[board.id].message_count = 0;
	UNLOCK(smp_block[board.id].lock);

	/* Now process the messages. */
	for (message = 0; message < message_count; message++)
	{
		switch (message_queue[message].message)
		{
			case SMP_HELP:
				copy_search_state(*sb, message_queue[message].data);
				break;
			case SMP_UPDATE:
				*sb = update(*sb, message_queue[message].data);
				break;
			case SMP_INSERT:
				insert(&board, message_queue[message].data, NULL);
				break;
			case SMP_UNSPLIT:
				*sb = unsplit(*sb, message_queue[message].data);
				break;
			case SMP_STOP:
				*sb = stop(*sb);
				break;
			case SMP_PARK:
				*sb = stop(*sb);
				(*sb)->search_state = SEARCH_PARK;
				smp_block[board.id].idle = TRUE;
				break;
		}
	}
}

/**
start_child_processors():
Makes all child processors enter the search function and start waiting for work.
Created 110107; last modified 110107
**/
void start_child_processors(void)
{
	int p;

	smp_copy_root(&smp_data->root_board, &board);
	for (p = 0; p < MAX_SPLIT_POINTS; p++)
		smp_copy_root(&split_point[p].board, &board);
	for (p = 1; p < process_count; p++)
		smp_tell(p, SMP_GO, 0);
}

/**
stop_child_processors():
Makes all child processors stop searching and wait for the next search.
Created 110107; last modified 110107
**/
void stop_child_processors(void)
{
	int p;

	for (p = 1; p < process_count; p++)
		smp_message(p, SMP_PARK, 0);
}

#ifdef ZCT_USE_SEMAPHORES
/**
lock():
Uses the given semaphore id to lock a section of code.
Created 082706; last modified 082706
**/
void lock(LOCK_T l)
{
	struct sembuf buf;

	buf.sem_num = l;
	buf.sem_op = -1;
	buf.sem_flg = 0;
	semop(sem_key, &buf, 1);
}

/**
unlock():
Uses the given semaphore id to unlock a section of code.
Created 082706; last modified 082706
**/
void unlock(LOCK_T l)
{
	struct sembuf buf;

	buf.sem_num = l;
	buf.sem_op = 1;
	buf.sem_flg = 0;
	semop(sem_key, &buf, 1);
}
#endif /* ZCT_USE_SEMAPHORES */

/**
smp_cleanup():
Takes care of all process-related stuff at exit, i.e. kill children, destroy shared memory
Created 080606; last modified 082606
**/
void smp_cleanup(void)
{
	int x;

	if (board.id == 0)
	{
		for (x = 1; x < process_count; x++)
		{
			print("Killing process %i(pid=%i)...\n", x, smp_block[x].pid);
			ExitProcess(smp_block[x].pid);
		}
#ifdef ZCT_USE_SEMAPHORES
		if (semctl(sem_key, 0, IPC_RMID) == -1)
			perror("semctl failed");
#endif
	}
}
/* This is just for compatibility with signal(). */
void smp_cleanup_sig(int x)
{
	smp_cleanup();
}

#endif /* SMP */
